// fg_bg_removal.c
// livido plugin
// (c) G. Finch (salsaman) 2005
//
// released under the GNU GPL 2
// see file COPYING or www.gnu.org for details

#include "plugins-livido.h"
#include <string.h>

// for parameters
#include "livido-lib.c"

livido_setup_f livido_setup;

livido_init_f removal_init;
livido_deinit_f removal_deinit;
livido_process_f removal_process;


typedef struct {
  uint8_t *av_luma_data;
  unsigned int av_count;
} static_data;

#define ABS(a)           (((a) < 0) ? -(a) : (a))

/* precomputed tables */
#define FP_BITS 16

static int Y_R[256];
static int Y_G[256];
static int Y_B[256];

static int Cb_R[256];
static int Cb_G[256];
static int Cb_B[256];
static int Cr_R[256];
static int Cr_G[256];
static int Cr_B[256];
static int conv_RY_inited = 0;


static int myround(double n)
{
  if (n >= 0) 
    return (int)(n + 0.5);
  else
    return (int)(n - 0.5);
}



static void init_RGB_to_YCbCr_tables(void)
{
  int i;

  /*
   * Q_Z[i] =   (coefficient * i
   *             * (Q-excursion) / (Z-excursion) * fixed-pogint-factor)
   *
   * to one of each, add the following:
   *             + (fixed-pogint-factor / 2)         --- for rounding later
   *             + (Q-offset * fixed-pogint-factor)  --- to add the offset
   *             
   */
  for (i = 0; i < 256; i++) {
    Y_R[i] = myround(0.299 * (double)i 
		     * 219.0 / 255.0 * (double)(1<<FP_BITS));
    Y_G[i] = myround(0.587 * (double)i 
		     * 219.0 / 255.0 * (double)(1<<FP_BITS));
    Y_B[i] = myround((0.114 * (double)i 
		      * 219.0 / 255.0 * (double)(1<<FP_BITS))
		     + (double)(1<<(FP_BITS-1))
		     + (16.0 * (double)(1<<FP_BITS)));

    Cb_R[i] = myround(-0.168736 * (double)i 
		      * 224.0 / 255.0 * (double)(1<<FP_BITS));
    Cb_G[i] = myround(-0.331264 * (double)i 
		      * 224.0 / 255.0 * (double)(1<<FP_BITS));
    Cb_B[i] = myround((0.500 * (double)i 
		       * 224.0 / 255.0 * (double)(1<<FP_BITS))
		      + (double)(1<<(FP_BITS-1))
		      + (128.0 * (double)(1<<FP_BITS)));

    Cr_R[i] = myround(0.500 * (double)i 
		      * 224.0 / 255.0 * (double)(1<<FP_BITS));
    Cr_G[i] = myround(-0.418688 * (double)i 
		      * 224.0 / 255.0 * (double)(1<<FP_BITS));
    Cr_B[i] = myround((-0.081312 * (double)i 
		       * 224.0 / 255.0 * (double)(1<<FP_BITS))
		      + (double)(1<<(FP_BITS-1))
		      + (128.0 * (double)(1<<FP_BITS)));
  }
  conv_RY_inited = 1;
}


static uint8_t 
calc_luma (uint8_t *pixel) {
  return (Y_R[pixel[0]] + Y_G[pixel[1]]+ Y_B[pixel[2]]) >> FP_BITS;
}

static uint32_t fastrand_val;

inline uint32_t fastrand(void)
{
#define rand_a 1073741789L
#define rand_c 32749L

  return (fastrand_val= rand_a * fastrand_val + rand_c);
}

void fastsrand(uint32_t seed)
{
  fastrand_val = seed;
}

static inline void 
make_black (uint8_t *pixel) {
  pixel[0]=pixel[1]=pixel[2]=(uint8_t)0;
}

/* declare our init function */
int common_init(livido_instance_t *inst) {
  static_data *sdata=(static_data *)((*inst->malloc) (sizeof(static_data)));

  sdata->av_luma_data=(uint8_t *)(*inst->malloc)(inst->in_channels->width*3*inst->in_channels->height*sizeof(uint8_t));
  sdata->av_count=0;
  inst->internal=sdata;

  //
  inst->num_in_parameters=1;
  inst->in_parameters=malloc(sizeof(livido_parameter_t));
  inst->in_parameters[0]=livido_init_int32_param("Threshold",64,0,255);
  //

  return LIVIDO_NO_ERROR;
}


static int common_process
(int type, livido_instance_t *inst, livido_frame_t *in, livido_frame_t *out) {
  // host should have returned one input channel and one output channel
  // (which can be NULL, since we exported the property INPLACE)

  // we also set our input/output channels same as, so we can assume both frame
  // sizes and palettes match

  register int i,j;
  int k,l;
  int rowstride_i=inst->in_channels->rowstrides[0];
  int rowstride_o;
  int width=inst->in_channels->width*3;
  int size=inst->in_channels->height*rowstride_i;

  uint8_t *old_pixel_data, *new_pixel_data;

  int inplace=(out==NULL);

  uint8_t luma;
  uint8_t av_luma;
  int32_t param1;
  uint8_t luma_threshold;

  static_data *sdata=(static_data *)(inst->internal);
  uint8_t *av_luma_data=(uint8_t *)(sdata->av_luma_data);

  old_pixel_data=in->pixel_data[0];

  livido_get_int32_prop(inst->in_parameters[0],"value",&param1);

  luma_threshold=param1&0xff;

  if (inplace) {
    new_pixel_data=old_pixel_data;
    rowstride_o=rowstride_i;
  }
  else {
    new_pixel_data=out->pixel_data[0];
    rowstride_o=inst->out_channels->rowstrides[0];      
  }
  sdata->av_count++;
 
  for (i=0;i<size;i+=rowstride_i) {
    l=i/rowstride_i*width;
    k=i/rowstride_i*rowstride_o;
    for (j=0;j<width-2;j+=3) {

      luma=calc_luma (&old_pixel_data[i+j]);
      if (sdata->av_count==1) {
	av_luma_data[l+j/3]=luma;
      }
      av_luma=(uint8_t)((double)(av_luma_data[l+j/3]*sdata->av_count)+(double)luma)/((double)sdata->av_count+1.);
      av_luma_data[l+j/3]=av_luma;
      if (ABS(luma-av_luma)>(luma_threshold)) {
	switch (type) {
	case 1:
	  // fire-ish effect
	  new_pixel_data[k+j]=((uint8_t)(((fastrand()&0xff)+(new_pixel_data[k+j+1]=((uint8_t) (fastrand()))))&0xff));     //R & G
	  new_pixel_data[k+j+2]=(uint8_t)0;                       //B
	  break;
	  //
	case 2:
	  break;
	case 0:
	  // make moving things black
	  make_black (&new_pixel_data[k+j]);
	}
      }
      else {
	// blue glow
	if (type==2) {
	  new_pixel_data[k+j]=new_pixel_data[k+j+1]=((uint8_t)(fastrand()&0xff));                                                 //R&G
	  new_pixel_data[k+j+2]=(uint8_t)255;  //B
	}}}}
  return LIVIDO_NO_ERROR;
}


int common_deinit (livido_instance_t *inst) {
  static_data *sdata;
  if (inst==NULL) return LIVIDO_ERROR_NULL_INSTANCE;
  sdata=(static_data *)(inst->internal);
  if (sdata!=NULL) {
    (*inst->free) (sdata->av_luma_data);
    (*inst->free) (sdata);
  }
  return LIVIDO_NO_ERROR; // success
}

int t1_process
(livido_instance_t *inst, livido_frame_t *in, livido_frame_t *out) {
  return common_process(0,inst,in,out);
}

int t2_process
(livido_instance_t *inst, livido_frame_t *in, livido_frame_t *out) {
  return common_process(1,inst,in,out);
}

int t3_process
(livido_instance_t *inst, livido_frame_t *in, livido_frame_t *out) {
  return common_process(2,inst,in,out);
}

// create 1 input and one matching output channel, no masks

// allow RGB888 and BGR888 palettes

livido_palette_t in_palettes[]={LIVIDO_PALETTE_BGR888,LIVIDO_PALETTE_RGB888,LIVIDO_PALETTE_END};
livido_palette_t out_palettes[]={LIVIDO_PALETTE_BGR888,LIVIDO_PALETTE_RGB888,LIVIDO_PALETTE_END};


static int paramdef1=200;
static int paramdef2=64;

//livido_parameter_t params1[]={{"amount","blend amount",LIVIDO_PARAM_NUMBER,0,0.,256.,&paramdef1,0,0},LIVIDO_PARAM_END};
//livido_parameter_t params2[]={{"threshold","threshold amount",LIVIDO_PARAM_NUMBER,0,0.,256.,&paramdef2,0,0},LIVIDO_PARAM_END};

livido_channel_template_t in_channels={"In 1",LIVIDO_CHANNEL_HOST_CAN_CHANGE_PALETTE,in_palettes,0,0,0,NULL};
livido_channel_template_t out_channels={"Out 1",LIVIDO_CHANNEL_SAME_AS_SIZE|LIVIDO_CHANNEL_SAME_AS_PALETTE|LIVIDO_CHANNEL_HOST_CAN_CHANGE_PALETTE,out_palettes,1,0,0,NULL};

livido_instance_template_t fx[]={{"fg_bg_removal_type1","salsaman","Foreground/background removal type 1",1,1.7,LIVIDO_PROPERTY_REALTIME|LIVIDO_PROPERTY_CAN_DO_INPLACE,NULL,NULL,&in_channels,&out_channels,common_init,common_deinit,t1_process,NULL,NULL,NULL},{"fg_bg_removal_type2","salsaman","Foreground/background removal type 2",1,1.7,LIVIDO_PROPERTY_REALTIME|LIVIDO_PROPERTY_CAN_DO_INPLACE,NULL,NULL,&in_channels,&out_channels,common_init,common_deinit,t2_process,NULL,NULL,NULL},{"fg_bg_removal_type3","salsaman","Foreground/background removal type 3",1,1.7,LIVIDO_PROPERTY_REALTIME|LIVIDO_PROPERTY_CAN_DO_INPLACE,NULL,NULL,&in_channels,&out_channels,common_init,common_deinit,t3_process,NULL,NULL,NULL}};

livido_instance_template_t *livido_setup(void) {
  init_RGB_to_YCbCr_tables();
  fx[0].next=&fx[1];
  fx[1].next=&fx[2];

  return fx;
}

