// simple_blend.c
// livido plugin
// (c) G. Finch (salsaman) 2005
//
// released under the GNU GPL 2
// see file COPYING or www.gnu.org for details

#include "plugins-livido.h"
#include "string.h"

// for parameters
#include "livido-lib.c"


livido_setup_f livido_setup;

livido_init_f simple_blend_init;
livido_deinit_f simple_blend_deinit;
livido_process_f simple_blend_process;


/* precomputed tables */
#define FP_BITS 16

static int Y_R[256];
static int Y_G[256];
static int Y_B[256];
static int Cb_R[256];
static int Cb_G[256];
static int Cb_B[256];
static int Cr_R[256];
static int Cr_G[256];
static int Cr_B[256];
static int conv_RY_inited = 0;

static int myround(double n)
{
  if (n >= 0) 
    return (int)(n + 0.5);
  else
    return (int)(n - 0.5);
}



static void init_RGB_to_YCbCr_tables(void)
{
  int i;

  /*
   * Q_Z[i] =   (coefficient * i
   *             * (Q-excursion) / (Z-excursion) * fixed-pogint-factor)
   *
   * to one of each, add the following:
   *             + (fixed-pogint-factor / 2)         --- for rounding later
   *             + (Q-offset * fixed-pogint-factor)  --- to add the offset
   *             
   */
  for (i = 0; i < 256; i++) {
    Y_R[i] = myround(0.299 * (double)i 
		     * 219.0 / 255.0 * (double)(1<<FP_BITS));
    Y_G[i] = myround(0.587 * (double)i 
		     * 219.0 / 255.0 * (double)(1<<FP_BITS));
    Y_B[i] = myround((0.114 * (double)i 
		      * 219.0 / 255.0 * (double)(1<<FP_BITS))
		     + (double)(1<<(FP_BITS-1))
		     + (16.0 * (double)(1<<FP_BITS)));

    Cb_R[i] = myround(-0.168736 * (double)i 
		      * 224.0 / 255.0 * (double)(1<<FP_BITS));
    Cb_G[i] = myround(-0.331264 * (double)i 
		      * 224.0 / 255.0 * (double)(1<<FP_BITS));
    Cb_B[i] = myround((0.500 * (double)i 
		       * 224.0 / 255.0 * (double)(1<<FP_BITS))
		      + (double)(1<<(FP_BITS-1))
		      + (128.0 * (double)(1<<FP_BITS)));

    Cr_R[i] = myround(0.500 * (double)i 
		      * 224.0 / 255.0 * (double)(1<<FP_BITS));
    Cr_G[i] = myround(-0.418688 * (double)i 
		      * 224.0 / 255.0 * (double)(1<<FP_BITS));
    Cr_B[i] = myround((-0.081312 * (double)i 
		       * 224.0 / 255.0 * (double)(1<<FP_BITS))
		      + (double)(1<<(FP_BITS-1))
		      + (128.0 * (double)(1<<FP_BITS)));
  }
  conv_RY_inited = 1;
}


inline uint8_t 
calc_luma (uint8_t *pixel) {
  return (Y_R[pixel[0]] + Y_G[pixel[1]]+ Y_B[pixel[2]]) >> FP_BITS;
}





/* declare our init function */
int common_init(livido_instance_t *inst) {
  //
  inst->num_in_parameters=1;
  inst->in_parameters=malloc(sizeof(livido_parameter_t));
  inst->in_parameters[0]=livido_init_int32_param("Blend amount",64,0,255);
  //
  return LIVIDO_NO_ERROR;
}


static int common_process
(int type, livido_instance_t *inst, livido_frame_t *in, livido_frame_t *out) {
  // host should have returned one input channel and one output channel
  // (which can be NULL, since we exported the property INPLACE)

  // we also set our input/output channels same as, so we can assume both frame
  // sizes and palettes match

  register int i,j;
  int k,l;
  int rowstride0=inst->in_channels->rowstrides[0];
  int rowstride1=inst->in_channels->next->rowstrides[0];
  int width=inst->in_channels->width*3;
  int size=inst->in_channels->height*rowstride0;

  int32_t bf;
  uint8_t blend_factor;

  // TODO
  unsigned char *old_pixel_data=in->pixel_data[0], *new_pixel_data, *blend_data=in->next->pixel_data[0];
  int inplace=(out==NULL);
  int rowstride2;

  livido_get_int32_prop(inst->in_parameters[0],"value",&bf);
  blend_factor=(uint8_t)bf;

  if (inplace) {
    new_pixel_data=old_pixel_data;
    rowstride2=rowstride0;
  }
  else {
    new_pixel_data=out->pixel_data[0];
    rowstride2=inst->out_channels->rowstrides[0];
  }
  for (i=0;i<size;i+=rowstride0) {
    k=i/rowstride0*rowstride1;
    l=i/rowstride0*rowstride2;
    for (j=0;j<width;j+=3) {
      switch (type) {
      case 0:
	// chroma blend
	new_pixel_data[l+j]=(uint8_t)((old_pixel_data[i+j]*blend_factor+blend_data[k+j]*(blend_factor^0xFF))>>8);
	new_pixel_data[l+j+1]=(uint8_t)((old_pixel_data[i+j+1]*blend_factor+blend_data[k+j+1]*(blend_factor^0xFF))>>8);
	new_pixel_data[l+j+2]=(uint8_t)((old_pixel_data[i+j+2]*blend_factor+blend_data[k+j+2]*(blend_factor^0xFF))>>8);
	break;
      case 1:
	if (calc_luma (&old_pixel_data[i+j])<(blend_factor)) {
	  new_pixel_data[l+j]=blend_data[k+j];
	  new_pixel_data[l+j+1]=blend_data[k+j+1];
	  new_pixel_data[l+j+2]=blend_data[k+j+2];
	}
	break;	
      case 2:
	// threshold luma blend (foreground)
	if (calc_luma (&blend_data[k+j])>(blend_factor)) {
	  new_pixel_data[l+j]=blend_data[k+j];
	  new_pixel_data[l+j+1]=blend_data[k+j+1];
	  new_pixel_data[l+j+2]=blend_data[k+j+2];
	}
	break;
      case 3:
	if (calc_luma (&old_pixel_data[i+j])>(blend_factor)) {
	  new_pixel_data[l+j]=blend_data[k+j];
	  new_pixel_data[l+j+1]=blend_data[k+j+1];
	  new_pixel_data[l+j+2]=blend_data[k+j+2];
	}
	break;	
      }
    }
  }
  return LIVIDO_NO_ERROR;
}


int common_deinit (livido_instance_t *inst) {
  return LIVIDO_NO_ERROR; // success
}

int chroma_process
(livido_instance_t *inst, livido_frame_t *in, livido_frame_t *out) {
  return common_process(0,inst,in,out);
}

int lumo_process
(livido_instance_t *inst, livido_frame_t *in, livido_frame_t *out) {
  return common_process(1,inst,in,out);
}

int lumu_process
(livido_instance_t *inst, livido_frame_t *in, livido_frame_t *out) {
  return common_process(2,inst,in,out);
}

int nlumo_process
(livido_instance_t *inst, livido_frame_t *in, livido_frame_t *out) {
  return common_process(3,inst,in,out);
}


// create 1 input and one matching output channel, no masks

// allow RGB888 and BGR888 palettes

livido_palette_t in_palettes[]={LIVIDO_PALETTE_RGB888,LIVIDO_PALETTE_BGR888,LIVIDO_PALETTE_END};
livido_palette_t out_palettes[]={LIVIDO_PALETTE_RGB888,LIVIDO_PALETTE_BGR888,LIVIDO_PALETTE_END};

livido_channel_template_t in_channels[]={{"foreground",LIVIDO_CHANNEL_HOST_CAN_RESIZE|LIVIDO_CHANNEL_HOST_CAN_CHANGE_PALETTE,in_palettes,0,0,0},{"background",LIVIDO_CHANNEL_HOST_CAN_RESIZE|LIVIDO_CHANNEL_HOST_CAN_CHANGE_PALETTE,in_palettes,0,0,0}};
livido_channel_template_t out_channels={"Out 1",LIVIDO_CHANNEL_SAME_AS_SIZE|LIVIDO_CHANNEL_SAME_AS_PALETTE|LIVIDO_CHANNEL_HOST_CAN_RESIZE|LIVIDO_CHANNEL_HOST_CAN_CHANGE_PALETTE,out_palettes,1,0,0};

					  //static int paramdef1=128;
					  //static int paramdef2=64;

					  //livido_parameter_t params1[]={{"amount","blend amount",LIVIDO_PARAM_NUMBER,0,0.,255.,&paramdef1,0,0},LIVIDO_PARAM_END};
					  //livido_parameter_t params2[]={{"threshold","threshold amount",LIVIDO_PARAM_NUMBER,0,0.,255.,&paramdef2,0,0},LIVIDO_PARAM_END};

livido_instance_template_t fx[]={{"chroma blend","salsaman","chroma blend",1,1.7,LIVIDO_PROPERTY_REALTIME|LIVIDO_PROPERTY_CAN_DO_INPLACE,NULL,NULL,in_channels,&out_channels,common_init,common_deinit,chroma_process,NULL,NULL,NULL},{"luma overlay","salsaman","luma overlay",1,1.7,LIVIDO_PROPERTY_REALTIME|LIVIDO_PROPERTY_CAN_DO_INPLACE,NULL,NULL,in_channels,&out_channels,common_init,common_deinit,lumo_process,NULL,NULL,NULL},{"luma underlay","salsaman","luma underlay",1,1.7,LIVIDO_PROPERTY_REALTIME|LIVIDO_PROPERTY_CAN_DO_INPLACE,NULL,NULL,in_channels,&out_channels,common_init,common_deinit,lumu_process,NULL,NULL,NULL},{"negative luma overlay","salsaman","nluma underlay",1,1.7,LIVIDO_PROPERTY_REALTIME|LIVIDO_PROPERTY_CAN_DO_INPLACE,NULL,NULL,in_channels,&out_channels,common_init,common_deinit,nlumo_process,NULL,NULL,NULL}};


livido_instance_template_t *livido_setup(void) {
  init_RGB_to_YCbCr_tables();
  fx[0].next=&fx[1];
  fx[1].next=&fx[2];
  fx[2].next=&fx[3];
  in_channels[0].next=&in_channels[1];
  return fx;
}

