// For official releases of livido.h, see www.dyne.org

// effects-livido.c
// LiVES
// (c) G. Finch 2004 <salsaman@xs4all.nl>
// Released under the GNU GPL 2.0
// see file COPYING or www.gnu.org for licensing details 

// Unless otherwise specified these licensing details apply to all other files
// distributed as part of LiVES


// TODO - block user from reconfiguring channels at inconvenient times (when this is implemented)

// TODO - update fps in instances when it changes

// TODO - handle channels with fixed sizes

#include "main.h"

#include "callbacks.h"
#include "support.h"
#include "dlfcn.h"
#include "rte_window.h"
#include "effects-livido.h"
#include "livido-lib.h"

#define MAX_LIVIDO_EFFECTS 1024

#define MAX_MODES_PER_KEY 8

#define MAX_LIVIDO_INSTANCES FX_KEYS

// dummy channel size when no clip is open
#define DEFAULT_CHANNEL_HSIZE 320
#define DEFAULT_CHANNEL_VSIZE 240

// fixed size for generators - TODO - make configurable
#define DEF_GEN_WIDTH 640
#define DEF_GEN_HEIGHT 480


// TODO - use a struct {
static livido_class_t *livido_fx[MAX_LIVIDO_EFFECTS];
static gchar *plugin_names[MAX_LIVIDO_EFFECTS];
static void *livido_handles[MAX_LIVIDO_EFFECTS];

// arrays of in/out channels, these get created when the plugins are loaded
// each pointer is to a linked list of channels// LiVES will always create these in the same order as the channel_templates, to assist with locating a particular channel_template
// these are internal copies linked to the template, when an instance is created we copy these to the inxtance channels

static livido_channel_t *my_in_channels[MAX_LIVIDO_EFFECTS];
static livido_channel_t *my_out_channels[MAX_LIVIDO_EFFECTS];


static livido_instance_t *livido_instance[MAX_LIVIDO_INSTANCES];
// }

// each 'hotkey' controls n instances, selectable as 'modes' or banks
static int key_to_instance[FX_KEYS>0?FX_KEYS-1:1][MAX_MODES_PER_KEY];
static int key_to_fx[FX_KEYS>0?FX_KEYS-1:1][MAX_MODES_PER_KEY];
static int key_modes[FX_KEYS>0?FX_KEYS-1:1];

static int next_free_instance=0;

// count of how many effects we have loaded
gint num_livido_effects=0;

// subtract this from num_livido_effects to give physical number of plugins
static gint num_modes;


// store keys so we now eg, which rte mask entry to xor when deiniting
static int fg_generator_key;
static int bg_generator_key;

// store modes too
static int fg_generator_mode;
static int bg_generator_mode;

// generators to start on playback, because of a problem in libvisual 
// - we must start generators after starting audio
static int bg_gen_to_start;
static int fg_gen_to_start;

// store the clip, this can sometimes get lost
static gint fg_generator_clip;

//#define DEBUG_LIVIDO


//////////////////////////////////////////// library/header functions




void livido_channels_free (livido_instance_t *inst) {
  livido_channel_t *channel,*channel_next;

  channel=inst->in_channels;

  while (channel!=NULL) {
    livido_properties_free(&channel->options);
    channel_next=channel->next;
    free(channel);
    channel=channel_next;
  }

  inst->in_channels=NULL;

  channel=inst->out_channels;

  while (channel!=NULL) {
    livido_properties_free(&channel->options);
    channel_next=channel->next;
    free(channel);
    channel=channel_next;
  }

  inst->out_channels=NULL;

}






livido_instance_t *livido_create_instance (livido_instance_template_t *class, livido_channel_t *inc, livido_channel_t *outc) {
  livido_instance_t *inst=g_malloc(sizeof(livido_instance_t));

  // optional - we could here check that inc and outc match 1 - 1 with templates

  inst->templ=class;
  inst->in_channels=inc;
  inst->out_channels=outc;
  inst->num_in_parameters=inst->num_out_parameters=0;
  inst->in_parameters=NULL;
  inst->out_parameters=NULL;
  inst->keyframe_next=NULL;
  inst->keyframe_prev=NULL;
  inst->fps=24.00;
  inst->scale_x=inst->scale_y=0.;
  inst->window_x=inst->window_y=inst->window_w=inst->window_h=0.;
  inst->threads=1;
  inst->malloc=&malloc;
  inst->free=&free;
  inst->memcpy=&memcpy;
  inst->memset=&memset;

  return inst;
}



void livido_free_instance (livido_instance_t *inst) {
  livido_channels_free(inst);

  // plugin should have done this in its deinit() function, but just in case:
  livido_parameters_free(inst);

  (*inst->free)(inst);
}



/////////////////////////////////////////////////////////////////////////////

static inline livido_channel_t *get_channel(livido_channel_t *chan, gint idx, gboolean count_disabled) {
  livido_channel_t *match=NULL;

  while (idx>0&&chan!=NULL) {
    if (count_disabled||!chan->flags&LIVIDO_CHANNEL_DISABLED) idx--;
    match=chan;
    chan=chan->next;
  }
  if (idx>0) return NULL;
  return match;
}



static inline gint enabled_in_channels (gint idx) {
  // template check
  livido_channel_t *chan=my_in_channels[idx];
  gint enabled=0;

  while (chan!=NULL) {
    if (!chan->flags&LIVIDO_CHANNEL_DISABLED) enabled++;
    chan=chan->next;
  }
  return enabled;
}

static inline gint enabled_out_channels (gint idx) {
  // template check
  livido_channel_t *chan=my_out_channels[idx];
  gint enabled=0;

  while (chan!=NULL) {
    if (!chan->flags&LIVIDO_CHANNEL_DISABLED) enabled++;
    chan=chan->next;
  }
  return enabled;
}


static inline gint ienabled_in_channels (gint idx) {
  // instance check
  livido_channel_t *chan;
  gint enabled=0;

  if (idx==-1||livido_instance[idx]==NULL) {
    g_printerr("WARN: tried to get in channels for non-existent instance !\n");
    return -1;
  }

  chan=livido_instance[idx]->in_channels;

  while (chan!=NULL) {
    if (!(chan->flags&LIVIDO_CHANNEL_DISABLED)) enabled++;
    chan=chan->next;
  }
  return enabled;
}

static inline gint ienabled_out_channels (gint idx) {
  // instance check
  livido_channel_t *chan;
  gint enabled=0;

  if (idx==-1||livido_instance[idx]==NULL) {
    g_printerr("WARN: tried to get out channels for non-existent instance !\n");
    return -1;
  }

  chan=livido_instance[idx]->out_channels;

  while (chan!=NULL) {
    if (!chan->flags&LIVIDO_CHANNEL_DISABLED) enabled++;
    chan=chan->next;
  }
  return enabled;
}


livido_channel_t *livido_channels_copy (livido_malloc_f *lmalloc, livido_channel_t *from_channels) {
  // return a newly malloc'ed copy of from_channels
  livido_channel_t *to_channels=NULL;
  livido_channel_t *prev_channel=NULL,*new_channel;

  int i;

  while (from_channels!=NULL) {
    new_channel=(*lmalloc)(sizeof(livido_channel_t));
    if (new_channel==NULL) return NULL; // return NULL on failure
    if (prev_channel!=NULL) prev_channel->next=new_channel;

    new_channel->templ=from_channels->templ;
    new_channel->flags=from_channels->flags;
    new_channel->palette=from_channels->palette;
    new_channel->width=from_channels->width;
    new_channel->palette=from_channels->palette;
    new_channel->shift_h=from_channels->shift_h;
    new_channel->shift_v=from_channels->shift_v;

    for (i=0;i<4;i++) {
      new_channel->rowstrides[i]=from_channels->rowstrides[i];
    }

    new_channel->options=NULL;   // TODO - copy options as well using livido_properties_copy
    new_channel->next=NULL;

    prev_channel=new_channel;
    if (to_channels==NULL) to_channels=new_channel;
    from_channels=from_channels->next;
  }
  
  return to_channels;
}





//////////////////////////////////////////////////////////////////////////////




static inline int get_next_free_instance(void) {
  int i;
  for (i=next_free_instance+1;i!=next_free_instance&&!(next_free_instance==-1&&i==MAX_LIVIDO_INSTANCES);i++) if (livido_instance[(i=(i==MAX_LIVIDO_INSTANCES?0:i))]==NULL) return i;
  return -1;
}


// check if palette is in the palette list for the channel template
static gboolean check_lvd_palette_list (livido_channel_template_t *chan_temp, livido_palette_t palette) {
  int i;

  if (chan_temp==NULL||chan_temp->name==NULL) return FALSE;
  if (palette==LIVIDO_PALETTE_END) return TRUE;

  for (i=0;chan_temp->palettes[i]!=LIVIDO_PALETTE_END;i++) {
    if (chan_temp->palettes[i]==palette) return TRUE;
  }
  return FALSE;
}


static void set_in_channel_palettes (gint idx, gint num_channels) {
  // set in channel palettes for filter[idx]
  // we also enable optional channels if we have to
  // in this case we fill first the mandatory channels,
  // then if necessary the optional channels
  livido_channel_t *channel;

  // our start state is with all optional channels disabled

  // fill mandatory channels

  for (channel=my_in_channels[idx];channel!=NULL;channel=channel->next) {
    if (check_lvd_palette_list ((livido_channel_template_t *)channel->templ,LIVIDO_PALETTE_BGR888)) channel->palette=LIVIDO_PALETTE_BGR888;
    else channel->palette=LIVIDO_PALETTE_RGB32;
    if (!channel->templ->flags&LIVIDO_CHANNEL_OPTIONAL) num_channels--; // mandatory channel
  }

  if (num_channels) {
    // OK we need to use some optional channels
    for (channel=my_in_channels[idx];channel!=NULL;channel=channel->next) {
      if (channel->templ->flags&LIVIDO_CHANNEL_OPTIONAL) {
	// got one !
	if (num_channels--) channel->flags^=LIVIDO_CHANNEL_DISABLED; // indicate we will use it
      }
    }
  }
}




static void set_out_channel_palettes (gint idx, gint num_channels) {
  // same as set_in_channel_palettes, but this time we cross ref the "same as"
  // palettes
  livido_channel_t *channel;

  for (channel=my_out_channels[idx];channel!=NULL;channel=channel->next) {
    if (check_lvd_palette_list ((livido_channel_template_t *)channel->templ,LIVIDO_PALETTE_BGR888)) channel->palette=LIVIDO_PALETTE_BGR888;
    else {
      if (check_lvd_palette_list ((livido_channel_template_t *)channel->templ,LIVIDO_PALETTE_RGB888)) channel->palette=LIVIDO_PALETTE_RGB888;
      else channel->palette=LIVIDO_PALETTE_RGB32;
    }
    if (!channel->templ->flags&LIVIDO_CHANNEL_OPTIONAL) num_channels--; // mandatory channel
  }

  if (num_channels) {
    // OK we need to use some optional channels
    for (channel=my_out_channels[idx];channel!=NULL;channel=channel->next) {
      if (channel->templ->flags&LIVIDO_CHANNEL_OPTIONAL) {
	// got one !
	if (num_channels--) channel->flags^=LIVIDO_CHANNEL_DISABLED; // indicate we will use it
      }
    }
  }

  // now we set same as palettes (Since we only use BGR888 now, this is irrelevant...)
  for (channel=my_in_channels[idx];channel!=NULL;channel=channel->next) {
    // in channels
    if (channel->templ->same_as>0&&channel->templ->flags&LIVIDO_CHANNEL_SAME_AS_PALETTE) channel->palette=get_channel(my_in_channels[idx],channel->templ->same_as,TRUE)->palette;
  }

  for (channel=my_out_channels[idx];channel!=NULL;channel=channel->next) {
    // out channels
    if (channel->templ->same_as>0&&channel->templ->flags&LIVIDO_CHANNEL_SAME_AS_PALETTE) channel->palette=get_channel(my_in_channels[idx],channel->templ->same_as,TRUE)->palette;
    else if (channel->templ->same_as>0&&channel->templ->flags&LIVIDO_CHANNEL_SAME_AS_PALETTE) channel->palette=get_channel(my_out_channels[idx],channel->templ->same_as,TRUE)->palette;
  }
  // same_size will be checked in init function
}




gboolean check_for_lives(livido_class_t *class, gint idx) {
  // for LiVES, currently:
  // all effects must take 0, 1 or 2 inputs and provide 
  // at least 1 output (for now)

  // all channels used must support BGR888 (for now)

  gint chans_in_mand=0; // number of mandatory channels
  gint chans_in_opt_max=0; // number of usable (by LiVES) optional channels
  gint chans_in_total=0;  // total number of channels
  gint chans_out_mand=0;
  gint chans_out_opt_max=0;
  gint chans_out_total=0;  // total number of channels
  gboolean is_generator;
  
  livido_channel_template_t *chtempl;
  livido_channel_t *in_channel,*out_channel,*prev_channel;

  my_in_channels[idx]=my_out_channels[idx]=NULL;

  // for now we will only load realtime effects

  if (!(class->flags&LIVIDO_PROPERTY_REALTIME)) return FALSE;

  // count number of mandatory and optional in_channels
  // now we can handle GBR888 and RGB32
  for (chtempl=class->in_channel_templates;chtempl!=NULL;chtempl=chtempl->next) {
    chans_in_total++;
    if (!chtempl->flags&LIVIDO_CHANNEL_OPTIONAL) {
      // is mandatory
      if (!check_lvd_palette_list (chtempl,LIVIDO_PALETTE_BGR888)&&!check_lvd_palette_list (chtempl,LIVIDO_PALETTE_RGB32)) return FALSE;
      chans_in_mand++;
    }
    else {
      // is optional
      if (check_lvd_palette_list (chtempl,LIVIDO_PALETTE_BGR888)||check_lvd_palette_list (chtempl,LIVIDO_PALETTE_RGB32)) chans_in_opt_max++;
    }
  }

  is_generator=!(chans_in_mand);

  // count number of mandatory and optional out_channels
  // restrict it just to BGR888 and RGB32, except for generators, which can return RGB888

  for (chtempl=class->out_channel_templates;chtempl!=NULL;chtempl=chtempl->next) {
    chans_out_total++;
    if (!chtempl->flags&LIVIDO_CHANNEL_OPTIONAL) {
      // is mandatory
      if (!check_lvd_palette_list (chtempl,LIVIDO_PALETTE_BGR888)&&!check_lvd_palette_list (chtempl,LIVIDO_PALETTE_RGB32)&&!(is_generator&&check_lvd_palette_list (chtempl,LIVIDO_PALETTE_RGB888))) return FALSE;
      chans_out_mand++;
    }
    else {
      // is optional
      if (check_lvd_palette_list (chtempl,LIVIDO_PALETTE_BGR888)||check_lvd_palette_list (chtempl,LIVIDO_PALETTE_RGB32)||(is_generator&&check_lvd_palette_list (chtempl,LIVIDO_PALETTE_RGB888))) chans_out_opt_max++;
    }
  }

  if (chans_in_mand>2) return FALSE; // we dont handle mixers yet...
  if (chans_out_mand>1||!(chans_out_mand+chans_out_opt_max)) return FALSE;

  // create internal copy of the channels for the template
  my_in_channels[idx]=(livido_channel_t *)g_malloc ((chans_in_total)*sizeof(livido_channel_t));
  my_out_channels[idx]=(livido_channel_t *)g_malloc ((chans_out_total)*sizeof(livido_channel_t));

  in_channel=my_in_channels[idx];
  prev_channel=NULL;

  for (chtempl=class->in_channel_templates;chtempl!=NULL;chtempl=chtempl->next) {
    in_channel->templ=chtempl;
    if (chtempl->flags&LIVIDO_CHANNEL_OPTIONAL) in_channel->flags=LIVIDO_CHANNEL_DISABLED;
    else in_channel->flags=0;
    in_channel->shift_h=in_channel->shift_v=0;
    in_channel->options=NULL;
    in_channel->next=NULL;
    if (prev_channel!=NULL) prev_channel->next=in_channel;
    prev_channel=in_channel++;
  }

  out_channel=my_out_channels[idx];
  prev_channel=NULL;

  for (chtempl=class->out_channel_templates;chtempl!=NULL;chtempl=chtempl->next) {
    out_channel->templ=chtempl;
    if (chtempl->flags&LIVIDO_CHANNEL_OPTIONAL) out_channel->flags=LIVIDO_CHANNEL_DISABLED;
    else out_channel->flags=0;
    out_channel->shift_h=out_channel->shift_v=0;
    out_channel->options=NULL;
    out_channel->next=NULL;
    if (prev_channel!=NULL) prev_channel->next=out_channel;
    prev_channel=out_channel++;
  }

  return TRUE;
}


// TODO - don't use fix sized channels (or allow resizing in apply)

void livido_init_all (void) {
// get list of plugins from directory and create our fx
  int i,j,idx=0,plugin_idx;
  gchar *plugin_name;
  GList *livido_fx_list=get_plugin_list(PLUGIN_LIVIDO_FX_BUILTIN,TRUE);
  gchar *plugin_path=g_strdup_printf("%s%s",PLUGIN_DIR,PLUGIN_LIVIDO_FX_BUILTIN);
  gchar *fx_path;
  gchar *msg;
  gint mode=0;
  gint skipped=0;
  gint key=-1;

  livido_setup_f *setup_fn;
  livido_class_t *fx_array;


#ifdef DEBUG_LIVIDO
  g_printerr("In livido init\n");
#endif
  num_modes=0;

  // danger Will Robinson !
  fg_gen_to_start=fg_generator_key=fg_generator_clip=fg_generator_mode=-1;
  bg_gen_to_start=bg_generator_key=bg_generator_mode=-1;

  // reset all instances to NULL
  for (i=0;i<MAX_LIVIDO_INSTANCES;livido_instance[i++]=NULL);
  for (i=0;i<FX_KEYS-1;i++) {
    key_modes[i]=0;
    for (j=0;j<MAX_MODES_PER_KEY;j++) {
      key_to_instance[i][j]=key_to_fx[i][j]=-1;
      
    }
  }

  // walk list and create fx structures
  for (plugin_idx=0;plugin_idx<g_list_length(livido_fx_list);plugin_idx++) {
    plugin_name=g_strdup(g_list_nth_data(livido_fx_list,plugin_idx));

    // TODO - replace with livido library fn
    fx_path=g_strdup_printf("%s/%s/%s",prefs->prefix_dir,plugin_path,plugin_name);

#ifdef DEBUG_LIVIDO
    g_printerr("Checking plugin %s\n",fx_path);
#endif

    if ((livido_handles[plugin_idx-skipped]=dlopen(fx_path,RTLD_NOW))) {
      setup_fn=dlsym(livido_handles[plugin_idx-skipped],"livido_setup");
      mode=0;
      if (!(fx_array=(*setup_fn)())) g_printerr("bang, plugin exploded !\n");
      else {
	for (;fx_array!=NULL&&idx<MAX_LIVIDO_EFFECTS;fx_array=fx_array->next) {
	  if (check_for_lives(fx_array,idx)) {
	    if (!mode) key++;

#ifdef DEBUG_LIVIDO
	    if (key<FX_KEYS-1) msg=g_strdup_printf("Loaded effect \"%s\" in plugin \"%s\"; assigned to key ctrl-%d, mode %d.\n",fx_array->name,plugin_name,key+1,mode+1);
	    else msg=g_strdup_printf("Loaded effect \"%s\" in plugin \"%s\", no key assigned\n",fx_array->name,plugin_name);
	    d_print(msg);
	    g_free(msg);
#endif

	    // we start with all optional channels disabled
	    if (enabled_in_channels(idx)>0) {
	      if (enabled_in_channels(idx)==1) {
		set_in_channel_palettes (idx,1);
	      }
	      else {
		set_in_channel_palettes (idx,2);
	      }
	    }
	    set_out_channel_palettes (idx,1);
	    livido_fx[idx]=fx_array;
	    plugin_names[idx]=g_strdup(plugin_name);
	    if (key<FX_KEYS-1) {
	      key_to_fx[key][mode]=idx;
	    }
	    mode++;
	    idx++;
	  }}}
      
      if (mode>0) num_modes+=--mode;
    }
    else {
      g_printerr(g_strdup_printf (_("Info: Unable to load plugin. %s\n"),fx_path));
      skipped++;
    }
    g_free(plugin_name);
  }
  num_livido_effects=idx-skipped;
  msg=g_strdup_printf(_ ("Loaded %d Livido effects\n"),idx-skipped);
  d_print(msg);
  g_free(msg);

  // TODO - add any rendered effects to fx submenu
}



GdkPixbuf *gdk_pixbuf_new_from_generator (livido_instance_t *inst, GError **error) {
  GdkPixbuf *pixbuf=NULL;
  guchar *pixels;
  livido_frame_t ofp;
  livido_timecode_t tc=0l;  // for now we pass in 0l as the timecode for out channel frames
  const livido_channel_t *out_channel;
  gint hsize,vsize,rowstride;

  if (inst==NULL) return NULL;
  
  if ((out_channel=get_channel(inst->out_channels,1,FALSE))==NULL) return NULL;

  // construct our gdk_pixbuf
  if ((pixbuf=gdk_pixbuf_new (GDK_COLORSPACE_RGB, FALSE, 8, (hsize=out_channel->width), (vsize=out_channel->height)))==NULL) return NULL;

  pixels=gdk_pixbuf_get_pixels (pixbuf);

  // WARNING WARNING !
  if ((rowstride=gdk_pixbuf_get_rowstride (pixbuf)!=out_channel->width*3)) return NULL;

  ofp.timecode=tc;
  ofp.pixel_data[0]=(void *)pixels;
  ofp.channel=out_channel;
  ofp.next=NULL;

  (*inst->templ->process)(inst,NULL,&ofp);
  
  if (out_channel->palette==LIVIDO_PALETTE_RGB888) {
    RGB2BGR_image (pixels,hsize,vsize,rowstride);
  }
  return pixbuf;
}

void livido_generator_end (livido_instance_t *inst) {
  // generator has stopped for one of the following reasons:
  // efect was de-inited; clip (bg/fg) was changed; playback stopped with fg
  gboolean is_bg=FALSE;
  gint current_file=mainw->current_file;

  if (inst==NULL) {
    g_printerr("  WARNING: inst was NULL !    ");
    return;
  }
  if (mainw->blend_file!=-1&&mainw->blend_file!=current_file&&mainw->files[mainw->blend_file]!=NULL&&mainw->files[mainw->blend_file]->ext_src==inst) is_bg=TRUE;

  if (!is_bg&&mainw->whentostop==STOP_ON_VID_END&&mainw->playing_file>0) {
    mainw->cancelled=1;
    return;
  }
  
  if (rte_window!=NULL) {
    if (!is_bg) {
      rtew_set_param_button (fg_generator_key,fg_generator_mode,FALSE);
      rtew_set_keych(fg_generator_key,FALSE);
    }
    else {
      rtew_set_param_button (bg_generator_key,bg_generator_mode,FALSE);
      rtew_set_keych(bg_generator_key,FALSE);
    }
  }

  if (!is_bg) close_pwindow(fg_generator_key,fg_generator_mode,FALSE);
  else close_pwindow(bg_generator_key,bg_generator_mode,FALSE);

  (*(inst->templ)->deinit)(inst);

  livido_free_instance(inst);

  if (is_bg) {
    livido_instance[key_to_instance[bg_generator_key][bg_generator_mode]]=NULL;
    key_to_instance[bg_generator_key][bg_generator_mode]=-1;
    if (mainw->rte&(1<<bg_generator_key)) mainw->rte^=(1<<bg_generator_key);
    bg_gen_to_start=bg_generator_key=bg_generator_mode=-1;
    mainw->current_file=mainw->blend_file;
    mainw->blend_file=-1;
  }
  else {
    livido_instance[key_to_instance[fg_generator_key][fg_generator_mode]]=NULL;
    key_to_instance[fg_generator_key][fg_generator_mode]=-1;
    if (mainw->rte&(1<<fg_generator_key)) mainw->rte^=(1<<fg_generator_key);
    fg_gen_to_start=fg_generator_key=fg_generator_clip=fg_generator_mode=-1;
    if (mainw->blend_file==mainw->current_file) mainw->blend_file=-1;
  }
  cfile->ext_src=NULL;
  cfile->disk_frames--;

  cfile->end=cfile->frames=cfile->disk_frames;

  if (cfile->frames!=2) {
    if (cfile->disk_frames<1) {
      if (mainw->playing_file==-1) {
	close_current_file();
      }
      else {
	system (g_strdup_printf("smogrify close %s",cfile->handle));
	if (cfile->perf_counter!=0) {
	  g_free(cfile->events[EVENT_FRAMES]);
	  g_free(cfile->events[EVENT_FPS]);
	}
	g_free(cfile);
	cfile=NULL;
	
	if (mainw->first_free_file==-1||mainw->first_free_file>mainw->current_file) {
	  mainw->first_free_file=mainw->current_file;
	}
      }

      if (!is_bg) {
	if (mainw->blend_file==mainw->current_file) mainw->blend_file=-1;
	if (mainw->playing_file>-1) {
	  mainw->reinit_effects=TRUE; // TODO - only do this if the frame size changes
	  if (mainw->blend_file>-1&&mainw->blend_file!=current_file) {
	    mainw->pre_src_file=mainw->blend_file;
	  }

	  if (mainw->pre_src_file!=-1&&mainw->files[mainw->pre_src_file]!=NULL) {
	    do_quick_switch (mainw->pre_src_file);
	  }
	  else prevclip_callback(NULL,NULL,0,0,NULL);

	  if (mainw->whentostop==STOP_ON_VID_END) {
	    mainw->cancelled=1;
	    return;
	  }
	}
      }
      else {
	mainw->current_file=current_file;
	if (mainw->playing_file==-1) {
	  switch_to_file ((mainw->current_file=0),current_file);
	}
      }
      return;
    }
  
    // some frames were recorded
    g_snprintf(cfile->name,256,g_strdup_printf(_ ("Untitled%d"),mainw->untitled_number++));
    set_menu_text (cfile->menuentry,cfile->name,FALSE);
    add_to_winmenu();
    cfile->is_untitled=TRUE;
  }

  cfile->start=1;

  // TODO - note start and end times of each recorded section
  // record event_frame and event_fps then use quantise_events to bring it
  // all in time

  //quantise_events(1); // TODO - debug memory problems with this

  cfile->clip_type=CLIP_TYPE_DISK;
  cfile->changed=TRUE;
  cfile->bpp=24;
  if (is_bg) mainw->current_file=current_file;
  else if (mainw->playing_file==-1) {
    switch_to_file ((mainw->current_file=0),current_file);
  }
  return;
}

void livido_bg_generator_end (livido_instance_t *inst) {
  // when we stop with a bg generator we want it to be restarted next time
  // i.e we will need a new clip for it

  int bg_gen_key=bg_generator_key;
  livido_generator_end (inst);
  bg_gen_to_start=bg_gen_key;
}


gboolean livido_generator_start (livido_instance_t *inst) {
  // make an "ephemeral clip"

  // cf. yuv4mpeg.c
  // start "playing" but receive frames from a plugin

  livido_channel_t *out_channel;

  gint old_file=mainw->current_file;

  // create a virtual clip
  gint new_file=mainw->first_free_file;

  if ((mainw->current_file>-1&&cfile!=NULL&&cfile->clip_type==CLIP_TYPE_GENERATOR)||(mainw->blend_file!=-1&&mainw->files[mainw->blend_file]!=NULL&&mainw->files[mainw->blend_file]->clip_type==CLIP_TYPE_GENERATOR)) return FALSE;

  if (mainw->is_processing&&!mainw->preview) return FALSE;
  if ((mainw->preview||(mainw->current_file>-1&&cfile!=NULL&&cfile->opening))&&!mainw->num_tr_applied) return FALSE;

  if (old_file==-1||cfile==NULL||cfile->frames>0) {
    if (!get_new_handle(new_file,g_strdup_printf (_ ("generator:%s"),inst->templ->name))) {
      mainw->error=TRUE;
      return FALSE;
    }

    if (mainw->playing_file==-1) mainw->pre_src_file=mainw->current_file;
    mainw->current_file=new_file;
    g_snprintf(cfile->type,40,"generator:%s",inst->templ->name);
    cfile->achans=0;
    cfile->asampsize=0;
    
    // open as a clip with 1 frame
    cfile->start=cfile->end=cfile->frames=1;
    cfile->arps=cfile->arate=0;
    cfile->changed=FALSE;
  }
  else {
    new_file=old_file;
    cfile->frames=2;
  }

  cfile->clip_type=CLIP_TYPE_GENERATOR;
  cfile->ext_src=inst;
 
  if (!mainw->num_tr_applied||mainw->current_file==-1) fg_generator_clip=new_file;

  cfile->disk_frames=0;
  cfile->pb_fps=cfile->fps=prefs->default_fps;

  // ignore disabled channels
  out_channel=get_channel(inst->out_channels,1,FALSE);

  cfile->hsize=out_channel->width;
  cfile->vsize=out_channel->height;

  cfile->opening=FALSE;
  cfile->proc_ptr=NULL;

  // allow clip switching
  cfile->is_loaded=TRUE;

  // record timecodes of frames - TODO - debug memory problems
  //cfile->perf_counter=cfile->perf_start=1;

  // if not playing, start playing
  if (mainw->playing_file==-1) {
    if (!mainw->num_tr_applied||old_file==-1) {
      switch_to_file((mainw->current_file=old_file),new_file);
      set_main_title(cfile->file_name,0);
      mainw->play_start=1;
      mainw->play_end=INT_MAX;
      if (mainw->num_tr_applied) {
	mainw->blend_file=mainw->current_file;
      }
    }
    else {
      mainw->blend_file=mainw->current_file;
      mainw->current_file=old_file;
      mainw->play_start=cfile->start;
      mainw->play_end=cfile->end;
      mainw->playing_sel=FALSE;
    }
    if (mainw->play_window!=NULL&&old_file==-1) {
      // usually preview or load_preview_frame would do this
      g_signal_handler_block(mainw->play_window,mainw->pw_exp_func);
    }


    // this is really horrible and hacky, it's just here as a quick fix to allow generators to be recorded
    // in future, starting a generator will not start playback, we will just set fg_gen_to_start/bg_gen_to_start
    play_file();

    if (fg_generator_clip!=-1) {
      if (rte_window!=NULL) rtew_set_keych(fg_generator_key,FALSE);
      inst=mainw->files[fg_generator_clip]->ext_src;
      if (mainw->blend_file==-1||mainw->blend_file==mainw->current_file||mainw->blend_file!=new_file) {
	mainw->current_file=fg_generator_clip;
	livido_generator_end (inst);
      }
    }
    if (mainw->pre_src_file>-1) {
      switch_to_file ((mainw->current_file=0),mainw->pre_src_file);
      mainw->pre_src_file=-2;
    }
  }
  else {
    // already playing
    mainw->current_file=old_file;
    if (!mainw->num_tr_applied||old_file==-1) {
      mainw->whentostop=NEVER_STOP;
      if (mainw->current_file==-1) mainw->current_file=new_file;
      do_quick_switch (new_file);
    }
    else {
      mainw->blend_file=new_file;
    }
  }
  return TRUE;
}






void set_channel_sizes (gint in_count,livido_channel_t *inc,livido_channel_t *outc) {

  if (in_count>0) {
    if (in_count>1) {
      if (mainw->current_file>-1&&cfile!=NULL) {
	// this is correct because we will always resize blend file to this
	inc->next->width=cfile->hsize;
	inc->next->height=cfile->vsize;
	inc->next->rowstrides[0]=3*cfile->hsize; // guess - TODO
      }
      else {
	inc->next->width=DEFAULT_CHANNEL_HSIZE;
	inc->next->height=DEFAULT_CHANNEL_VSIZE;
	inc->next->rowstrides[0]=DEFAULT_CHANNEL_HSIZE*3;
      }
    }

    if (mainw->current_file>-1&&cfile!=NULL) {
      inc->width=cfile->hsize;
      inc->height=cfile->vsize;
      inc->rowstrides[0]=3*cfile->hsize;
    }
    else {
      mainw->reinit_effects=TRUE;
      inc->width=DEFAULT_CHANNEL_HSIZE;
      inc->height=DEFAULT_CHANNEL_VSIZE;
      inc->rowstrides[0]=DEFAULT_CHANNEL_HSIZE*3;
    }
      
    if (mainw->current_file>-1&&cfile!=NULL) {
      outc->width=cfile->hsize;
      outc->height=cfile->vsize;
      outc->rowstrides[0]=cfile->hsize*3;
    }
    else {
      outc->width=DEFAULT_CHANNEL_HSIZE;
      outc->height=DEFAULT_CHANNEL_VSIZE;
      outc->rowstrides[0]=DEFAULT_CHANNEL_HSIZE*3;
    }
  }
  else {
    outc->width=DEF_GEN_WIDTH;
    outc->height=DEF_GEN_HEIGHT;
    outc->rowstrides[0]=outc->width*3;
  }
}


gboolean livido_key_valid (int key) {
  if (key_to_fx[key][key_modes[key]]==-1) return FALSE;
  return TRUE;
}

gboolean livido_keymode_valid (gint key, gint mode) {
  if (key_to_fx[--key][mode]==-1) return FALSE;
  return TRUE;
}

gint livido_keymode_get_idx(gint key, gint mode) {
  return (key_to_fx[--key][mode]);
}



gboolean livido_init_effect(int hotkey, gint rowstride) {
  // mainw->osc_block should be set to TRUE before calling this function !
  livido_class_t *fx;
  livido_instance_t *new_instance;
  livido_channel_t *inc,*outc;
  gint num_tr_applied;
  gboolean fg_modeswitch=FALSE,is_trans=FALSE,gen_start=FALSE,is_resize=FALSE;
  gint rte_keys=mainw->rte_keys;
  gint inc_count;

  if (hotkey<0) {
    is_resize=TRUE;
    hotkey=-hotkey-1;
  }

  if (hotkey==fg_generator_key) {
    fg_modeswitch=TRUE;
  }

  if (!livido_key_valid (hotkey)) {
    return FALSE;
  }

  if (next_free_instance==-1) {
    d_print(_ ("Warning - no more effect slots left !\n"));
    return FALSE;
  }

  inc_count=enabled_in_channels(key_to_fx[hotkey][key_modes[hotkey]]);

  // TODO - block template channel changes
  // we must stop any old generators
  // warning = in future, bg *and* fg generators could change
  if (mainw->num_tr_applied&&mainw->blend_file!=-1&&mainw->blend_file!=mainw->current_file&&mainw->files[mainw->blend_file]!=NULL&&mainw->files[mainw->blend_file]->clip_type==CLIP_TYPE_GENERATOR) {
    if (bg_gen_to_start==-1) {
      livido_generator_end(mainw->files[mainw->blend_file]->ext_src);
    }
    bg_gen_to_start=bg_generator_key=bg_generator_mode=-1;
    mainw->blend_file=-1;
  }
  if (mainw->current_file>0&&cfile->clip_type==CLIP_TYPE_GENERATOR&&fg_modeswitch) {
    if (mainw->noswitch||mainw->is_processing||mainw->preview) return FALSE; // stopping fg gen will cause clip to switch
    if (mainw->playing_file>-1&&mainw->whentostop==STOP_ON_VID_END&&inc_count!=0) {
      mainw->cancelled=1;
    }
    else {
      livido_generator_end (cfile->ext_src);
      mainw->reinit_effects=TRUE;
      fg_generator_key=fg_generator_clip=fg_generator_mode=-1;
    }
  }

  fx=livido_fx[key_to_fx[hotkey][key_modes[hotkey]]];

  if (inc_count==2) {
    mainw->num_tr_applied++; // increase trans count
    if (mainw->num_tr_applied==1&&!is_resize) {
      mainw->blend_file=mainw->current_file;
      mainw->blend_file_step=(mainw->current_file>-1&&cfile->pb_fps<0)?-1:1;
    }
  }
  else if (inc_count==0) {
    // aha - a generator
     if (mainw->playing_file==-1) {
      // if we are not playing, we will postpone creating the instance
      // this is a workaround for a problem in libvisual
      if (mainw->num_tr_applied&&mainw->current_file!=-1) {
	bg_gen_to_start=hotkey;
      }
      else fg_gen_to_start=hotkey;
      gen_start=TRUE;
    }
     else if (!fg_modeswitch&&!mainw->num_tr_applied&&(mainw->noswitch||mainw->is_processing||mainw->preview)) return FALSE;
  }

  // create copy channels for the instance
  inc=livido_channels_copy ((livido_malloc_f *)&g_try_malloc,my_in_channels[key_to_fx[hotkey][key_modes[hotkey]]]);
  outc=livido_channels_copy ((livido_malloc_f *)&g_try_malloc,my_out_channels[key_to_fx[hotkey][key_modes[hotkey]]]);
  set_channel_sizes (inc_count,inc,outc);

  new_instance=livido_create_instance(fx,inc,outc);
  // TODO - unblock template channel changes

  //new_instance->malloc=(livido_malloc_f *)&g_try_malloc;
  //new_instance->free=(livido_free_f *)&g_free;
  
  livido_instance[next_free_instance]=new_instance;
  key_to_instance[hotkey][key_modes[hotkey]]=next_free_instance;

  if (!gen_start) {
    if ((*fx->init)(new_instance)) {
      d_print (g_strdup_printf (_ ("Failed to start instance %s\n"),new_instance->templ->name));
      livido_free_instance(new_instance);
      livido_instance[key_to_instance[hotkey][key_modes[hotkey]]]=NULL;
      key_to_instance[hotkey][key_modes[hotkey]]=-1;
      if (is_trans) mainw->num_tr_applied--;
      return FALSE;
    }
  }
  // TODO - check param->template mapping

  if (inc_count==0) {
    // generator start
    if (mainw->num_tr_applied&&!fg_modeswitch&&mainw->current_file>-1) {
      // transition is on, make into bg clip
      bg_generator_key=hotkey;
      bg_generator_mode=key_modes[hotkey];
    }
    else {
      // no transition, make fg (or kb was grabbed for fg generator)
      fg_generator_key=hotkey;
      fg_generator_mode=key_modes[hotkey];
    }

    // start the generator and maybe start playing
    num_tr_applied=mainw->num_tr_applied;
    if (fg_modeswitch) mainw->num_tr_applied=0; // force to fg

    if (!livido_generator_start(new_instance)) {
      d_print (g_strdup_printf (_ ("Unable to start generator %s\n"),new_instance->templ->name));
      if (mainw->num_tr_applied&&mainw->current_file>-1) {
	bg_gen_to_start=bg_generator_key=bg_generator_mode=-1;
      }
      else {
	fg_generator_key=fg_generator_clip=fg_generator_mode=-1;
      }
      if (fg_modeswitch) mainw->num_tr_applied=num_tr_applied;
      return FALSE;
    }
    // TODO - problem if modeswitch triggers playback
    if (fg_modeswitch) mainw->num_tr_applied=num_tr_applied;
    if (fg_generator_key!=-1) mainw->rte|=(1<<fg_generator_key);
    if (bg_generator_key!=-1) {
      mainw->rte|=(1<<bg_generator_key);
      if (rte_window!=NULL) rtew_set_keych(bg_generator_key,TRUE);
    }
  }

  if (rte_keys==hotkey) mainw->rte_keys=rte_keys;
  if (rte_window!=NULL) rtew_set_param_button (hotkey,key_modes[hotkey],TRUE);

  next_free_instance=get_next_free_instance();
  return TRUE;
}



gboolean livido_playback_gen_start (void) {
  // init generators on pb. We have to do this after audio startup
  livido_instance_t *inst;
  int error=LIVIDO_NO_ERROR;

  mainw->osc_block=TRUE;
  if (bg_gen_to_start!=-1) {
    if (fg_gen_to_start==-1) {
      if (enabled_in_channels(key_to_fx[bg_gen_to_start][key_modes[bg_gen_to_start]])==0) { // check is still gen
	inst=livido_instance[key_to_instance[bg_gen_to_start][key_modes[bg_gen_to_start]]];
	if (inst==NULL) {
	  // 2nd playback
	  gint playing_file=mainw->playing_file;
	  mainw->playing_file=-100; //kludge to stop playing a second time
	  if (livido_init_effect (bg_gen_to_start,cfile->hsize*3)==FALSE) {
	    error++;
	  }
	  mainw->playing_file=playing_file;
	  inst=livido_instance[key_to_instance[bg_gen_to_start][key_modes[bg_gen_to_start]]];
	}
	else error=(*inst->templ->init)(inst);
	if (error!=LIVIDO_NO_ERROR) {
	  d_print (g_strdup_printf (_ ("Failed to start generator %s\n"),inst->templ->name));
	  if (inst!=NULL) {
	    (*inst->templ->deinit)(inst);
	    livido_free_instance(inst);
	  }
	  livido_instance[key_to_instance[bg_gen_to_start][key_modes[bg_gen_to_start]]]=NULL;
	  key_to_instance[bg_gen_to_start][key_modes[bg_gen_to_start]]=-1;
	  bg_gen_to_start=-1;
	  if (mainw->rte&(1<<bg_gen_to_start)) mainw->rte^=(1<<bg_gen_to_start);
	  mainw->osc_block=FALSE;
	  return FALSE;
	  // TODO - check param->template mapping
	}
	mainw->files[mainw->blend_file]->ext_src=inst;
	next_free_instance=get_next_free_instance();
	if (rte_window!=NULL) rtew_set_param_button (bg_gen_to_start,key_modes[bg_gen_to_start],TRUE);
      }
    }
    bg_gen_to_start=-1;
  }

  if (fg_gen_to_start!=-1) {
    if (enabled_in_channels(key_to_fx[fg_gen_to_start][key_modes[fg_gen_to_start]])==0) { // check is still gen
      inst=livido_instance[key_to_instance[fg_gen_to_start][key_modes[fg_gen_to_start]]];
      if (inst!=NULL&&(*inst->templ->init)(inst)) {
	d_print (g_strdup_printf (_ ("Warn: Failed to start generator %s\n"),inst->templ->name));
	(*inst->templ->deinit)(inst);
	livido_free_instance(inst);
	livido_instance[key_to_instance[fg_gen_to_start][key_modes[fg_gen_to_start]]]=NULL;
	key_to_instance[fg_gen_to_start][key_modes[fg_gen_to_start]]=-1;
	fg_gen_to_start=-1;
	cfile->ext_src=NULL;
	mainw->osc_block=FALSE;
	return FALSE;
      }
      // TODO - check param->template mapping
      cfile->ext_src=inst;
      next_free_instance=get_next_free_instance();
      if (mainw->save_to_ext) toggle_record();
      if (rte_window!=NULL) rtew_set_param_button (fg_gen_to_start,key_modes[fg_gen_to_start],TRUE);
    }
    fg_gen_to_start=-1;
  }

  mainw->osc_block=FALSE;
  return TRUE;
}
 




#define KEYSCALE 256.

gint livido_get_blend_factor(int hotkey) {
  // mainw->osc_block should be set to TRUE before calling this function !

  livido_instance_t *instance;
  int idx;
  int32_t ptype,vali,mini,maxi;

  if (hotkey<0) return 0;
  idx=key_to_instance[hotkey][key_modes[hotkey]];

  if (idx==-1||(instance=livido_instance[idx])==NULL) return 0;

  if (instance->num_in_parameters==0) return 0;

  livido_get_int32_prop (instance->in_parameters[0],"type",&ptype);

  switch (ptype) {
  case LIVIDO_PORT_INT32:
    livido_get_int32_prop (instance->in_parameters[0],"value",&vali);
    livido_get_int32_prop (instance->in_parameters[0],"min",&mini);
    livido_get_int32_prop (instance->in_parameters[0],"max",&maxi);
    return (gint)((gdouble)(vali-mini)/(gdouble)(maxi-mini)*KEYSCALE+.5);
  }

  return 0;
}

void livido_set_blend_factor(int hotkey, gint blend_factor) {
  // mainw->osc_block should be set to TRUE before calling this function !
  livido_instance_t *instance;
  int idx=key_to_instance[hotkey][key_modes[hotkey]];
  int32_t ptype,vali,mini,maxi;
  GList *list=NULL;

  if (hotkey<0) return;
  idx=key_to_instance[hotkey][key_modes[hotkey]];

  if (idx==-1||(instance=livido_instance[idx])==NULL) return;

  if (instance->num_in_parameters==0) return;

  livido_get_int32_prop (instance->in_parameters[0],"type",&ptype);
  
  switch (ptype) {
  case LIVIDO_PORT_INT32:
    livido_get_int32_prop (instance->in_parameters[0],"value",&vali);
    livido_get_int32_prop (instance->in_parameters[0],"min",&mini);
    livido_get_int32_prop (instance->in_parameters[0],"max",&maxi);
    livido_set_int32_prop(instance->in_parameters[0],"value",(int32_t)((gdouble)mini+((gdouble)blend_factor/KEYSCALE*(gdouble)(maxi-mini))+.5));
    livido_get_int32_prop (instance->in_parameters[0],"value",&vali);

    list=g_list_append(list,g_strdup_printf("%d",vali));
    list=g_list_append(list,g_strdup_printf("%d",mini));
    list=g_list_append(list,g_strdup_printf("%d",maxi));
    update_pwindow(hotkey,0,list);
    g_list_free(list);

    break;
  }

}

void livido_deinit_effect(int hotkey) {
  // mainw->osc_block should be set before calling this function !
  // caller should also handle mainw->rte

  const livido_class_t *fx;
  livido_instance_t *instance;
  gboolean is_resize=FALSE;
  gboolean was_transition=FALSE;
  gint num_in_chans;
  gint idx;

  if (hotkey<0) {
    is_resize=TRUE;
    hotkey=-hotkey-1;
  }

  if (hotkey>FX_KEYS-1) return;

  if ((idx=key_to_instance[hotkey][key_modes[hotkey]])==-1) return;
  
  instance=livido_instance[idx];
  if (instance==NULL) return;

  num_in_chans=ienabled_in_channels(idx);

  if (num_in_chans==0) {
    // is generator
    if (mainw->playing_file>-1&&mainw->whentostop==STOP_ON_VID_END) {
      mainw->cancelled=1;
    }
    else {
      livido_generator_end (instance);
    }
    return;
  }
  else {
    if (rte_window!=NULL) rtew_set_param_button (hotkey,key_modes[hotkey],FALSE);
    close_pwindow(hotkey,key_modes[hotkey],is_resize);
    fx=instance->templ;
    (*fx->deinit)(instance);
  }

  if (num_in_chans==2) {
    was_transition=TRUE;
    mainw->num_tr_applied--;
  }

  livido_free_instance(instance);

  livido_instance[idx]=NULL;
  key_to_instance[hotkey][key_modes[hotkey]]=-1;

  if (was_transition&&!is_resize) {
    if (mainw->num_tr_applied<1) {
      if (bg_gen_to_start!=-1) bg_gen_to_start=-1;
      if (mainw->blend_file!=-1&&mainw->blend_file!=mainw->current_file&&mainw->files[mainw->blend_file]!=NULL&&mainw->files[mainw->blend_file]->clip_type==CLIP_TYPE_GENERATOR) {
	// all transitions off, so end the bg generator
	livido_generator_end (mainw->files[mainw->blend_file]->ext_src);
      }
      mainw->blend_file=-1;
    }
  }
}


void **store_livido_param_values (gint key) {
  // TODO - eventually use store_rfx_param_values, etc.
  // requires stabilisation of livido parameters
  livido_instance_t *inst;
  int i;
  void **store=NULL;
  int type;

  if (key_to_instance[key][key_modes[key]]==-1||(inst=livido_instance[key_to_instance[key][key_modes[key]]])==NULL) return NULL;

  if (inst->num_in_parameters==0) return NULL;
 
  // TODO - need a livido_get_atom_size(int atom_type);

  store=(void **)g_malloc ((inst->num_in_parameters)*(sizeof(void *)));

  for (i=0;i<inst->num_in_parameters;i++) {
    livido_get_int32_prop(inst->in_parameters[i],"type",&type);
    switch (type) {
    case LIVIDO_PORT_SWITCH:
	store[i]=g_malloc (sizeof(int));
	livido_get_boolean_prop(inst->in_parameters[i],"value",store[i]);
	break;
    case LIVIDO_PORT_INT32:
    case LIVIDO_PORT_STRING_LIST:
	store[i]=g_malloc (sizeof(int));
	livido_get_int32_prop(inst->in_parameters[i],"value",store[i]);
	break;
    default:
      store[i]=NULL;
    }
  }
  return store;
}


void set_livido_params_from_store (gint key, void **store) {
  livido_instance_t *inst;
  int i;
  int type;

  GList *list;

  if (store==NULL||key_to_instance[key][key_modes[key]]==-1||(inst=livido_instance[key_to_instance[key][key_modes[key]]])==NULL) {
    return;
  }
  if (inst->num_in_parameters==0) {
    return;
  }

  for (i=0;i<inst->num_in_parameters;i++) {
    list=NULL;
    livido_get_int32_prop(inst->in_parameters[i],"type",&type);
    switch (type) {
    case LIVIDO_PORT_SWITCH:
      {
	int *int_p=(int *)store[i];
	livido_set_boolean_prop(inst->in_parameters[i],"value",*int_p);
	list=g_list_append(list,g_strdup_printf("%d",*int_p));
	update_pwindow(key,i,list);
	g_list_free(list);
	break;
      }
    case LIVIDO_PORT_INT32:
      {
	int min,max;
	int *int_p=(int *)store[i];
	livido_set_int32_prop(inst->in_parameters[i],"value",*int_p);

	livido_get_int32_prop(inst->in_parameters[i],"min",&min);
	livido_get_int32_prop(inst->in_parameters[i],"max",&max);
	list=g_list_append(list,g_strdup_printf("%d",*int_p));
	list=g_list_append(list,g_strdup_printf("%d",min));
	list=g_list_append(list,g_strdup_printf("%d",max));
	update_pwindow(key,i,list);
	g_list_free(list);
	break;
      }
    case LIVIDO_PORT_STRING_LIST:
      {
	gchar **string_list;
	int *int_p=(int *)store[i];
	
	livido_set_int32_prop(inst->in_parameters[i],"value",*int_p);
	string_list=(char **)g_malloc((livido_get_num_elements(inst->in_parameters[i],"list_values")+1)*sizeof(char *));
	livido_get_string_array(inst->in_parameters[i],"list_values",string_list);
	list=g_list_append(list,g_strdup_printf("\"%s\"",subst (string_list[*int_p],"\"","\\\"")));
	update_pwindow(key,i,list);
	g_strfreev(string_list);
	g_list_free(list);
	break;
      }
    default:
      break;
    }
  }
}



void livido_params_store_free (gint key, void **store) {
  livido_instance_t *inst;
  int i;

  if (store==NULL||key_to_instance[key][key_modes[key]]==-1||(inst=livido_instance[key_to_instance[key][key_modes[key]]])==NULL) {
    return;
  }
  for (i=0;i<inst->num_in_parameters;i++) {
    if (store[i]!=NULL) g_free (store[i]);
  }
}



void livido_init_resize(gint rowstride) {
  // this is called a) at the start of playback, and b) when the frame size of
  // the current file changes
  void **param_store=NULL;
  int i;
  gint blend_file=mainw->blend_file;
  guint rte_keys=mainw->rte_keys;

  livido_channel_t *first_in,*first_out;
  livido_instance_t *instance;
  gint idx;

  mainw->osc_block=TRUE;

  for (i=0;i<FX_KEYS-1&&i<num_livido_effects;i++) {
    if (mainw->rte&(1<<i)) {
      // re-init only active effects
      if ((idx=key_to_instance[i][key_modes[i]])==-1) continue;
      instance=livido_instance[idx];
      if (instance==NULL) continue;
      if (ienabled_in_channels(idx)>0) {
	// and don't re-init generators

	// TODO - check all channels
	first_in=get_channel(instance->in_channels,1,FALSE);
	first_out=get_channel(instance->out_channels,1,FALSE);

	// don't resize any effects where the first enabled in_channel and first enabled out_channel are both resizable
	if ((first_in!=NULL&&!(first_in->templ->flags&LIVIDO_CHANNEL_HOST_CAN_RESIZE))||(first_out!=NULL&&!(first_out->templ->flags&LIVIDO_CHANNEL_HOST_CAN_RESIZE))) {
	  param_store=store_livido_param_values (i);
	  livido_deinit_effect(-i-1);
	  mainw->rte_keys=rte_keys;
	  // reset all channel sizes
	  if (livido_init_effect(-i-1,rowstride)) {
	    restore_pwindow(i);
	    if (param_store!=NULL) {
	      set_livido_params_from_store (i,param_store);
	    }
	  }
	  else if (mainw->rte&(1<<i)) mainw->rte^=(1<<i);
	  if (param_store!=NULL) {
	    livido_params_store_free (i,param_store);
	    g_free (param_store);
	  }
	}
	else {
	  // just reset the channel sizes
	  set_channel_sizes (ienabled_in_channels(idx),instance->in_channels,instance->out_channels);
	}
      }
    }
  }
  mainw->blend_file=blend_file;
  mainw->osc_block=FALSE;
}




void livido_apply_effects(int height, int rowstride, guchar **in_pixel_data, livido_palette_t in_palette, livido_timecode_t tc) {
  // for now, the in_palette will always come in as BGR888, this can be changed back and forth, but the final output should be BGR888 again
  
  // there could be a small optimisation, if mainw->ext_playback is TRUE
  // then the final output could be left as mainw->ext_palette
  // sometimes this can be UYVY

  // palettes can now be converted between effects, optimisation of in/out palettes has not been done yet - TODO
  // livido also allows some channels to switch palette during processing, we could do this also to optimise



  // for now, we also use the same timecode for all in frames (foreground and background)
  // for realtime playback, this is counted from play start
  // - for rendering, we set the timecode to framecount/fps

  // timecode of out frames is set to 0l


  // TODO ** - handle case where a channel is of fixed size (can be set by plugin)

  // TODO - after calling process(), update the param window with any values which changed.


  int i;

  guchar *orig_pixel_data_in[2];
  guchar *conv_pixel_data=NULL;

  const livido_class_t *fx;
  livido_instance_t *instance;

  livido_frame_t in_frames[2];
  livido_frame_t out_frames[1];
  
  orig_pixel_data_in[0]=NULL;
  orig_pixel_data_in[1]=NULL;

  in_frames[0].timecode=in_frames[1].timecode=tc;
  in_frames[0].pixel_data[0]=in_pixel_data[0];
  in_frames[0].next=in_frames[1].next=NULL;

  if (in_pixel_data[1]!=NULL) {
    // we have two frame sources
    in_frames[1].pixel_data[0]=in_pixel_data[1];
  }

  for (i=0;i<FX_KEYS-1;i++) {
    if (livido_key_valid(i)) {
      if (mainw->rte&(1<<i)) {
	mainw->osc_block=TRUE;
	if (key_to_instance[i][key_modes[i]]==-1) continue;
	instance=livido_instance[key_to_instance[i][key_modes[i]]];
	if (instance==NULL) continue;

	in_frames[0].channel=get_channel(instance->in_channels,1,FALSE);
	if (in_frames[0].channel==NULL) continue;  // ignore generators

	in_frames[1].channel=get_channel(instance->in_channels,2,FALSE);
	if (in_frames[1].channel!=NULL) {
	  if (in_pixel_data[1]==NULL) continue;  // second frame is missing, skip this transition
	  in_frames[0].next=&in_frames[1];  // transition
	}
	else in_frames[0].next=NULL;

	fx=instance->templ;

	if (!(fx->flags&LIVIDO_PROPERTY_CAN_DO_INPLACE)) {
	  // plugin cannot do inplace
	  out_frames[0].channel=get_channel(instance->out_channels,1,FALSE);  // ignore disabled channels
	  out_frames[0].timecode=0l;

	  // do palette conversions as necessary

	  if (in_frames[0].channel->palette==LIVIDO_PALETTE_RGB32) {
	    // convert in_frame BGR888-->RGB32
	    orig_pixel_data_in[0]=in_frames[0].pixel_data[0];
	    in_frames[0].pixel_data[0]=convert_frame_palette(in_frames[0].pixel_data[0], rowstride, height, LIVIDO_PALETTE_BGR888, LIVIDO_PALETTE_RGB32, NULL);
	  }
	  if (in_frames[1].channel!=NULL&&in_frames[1].channel->palette==LIVIDO_PALETTE_RGB32) {
	    // convert in_frame BGR888-->RGB32
	    orig_pixel_data_in[1]=in_frames[1].pixel_data[0];
	    in_frames[1].pixel_data[0]=convert_frame_palette(in_frames[1].pixel_data[0], rowstride, height, LIVIDO_PALETTE_BGR888, LIVIDO_PALETTE_RGB32, NULL);
	  }

	  if (out_frames[0].channel->palette==LIVIDO_PALETTE_RGB32) {
	    conv_pixel_data=out_frames[0].pixel_data[0]=g_malloc(rowstride/3*height*4*sizeof(guchar));
	  }
	  else {
	    out_frames[0].pixel_data[0]=g_malloc(out_frames[0].channel->height*out_frames[0].channel->rowstrides[0]);
	  }

	  (*fx->process)(instance,in_frames,out_frames);	    

	  // TODO - handle case where we need to resize the out channel data. This should only happen if we have a plugin with
	  // a fixed output size

	  if (out_frames[0].channel->palette==LIVIDO_PALETTE_RGB32) {
	    // convert out_frame RGB32 --> BGR888
	    convert_frame_palette(conv_pixel_data, rowstride/3*4, height, LIVIDO_PALETTE_RGB32, LIVIDO_PALETTE_BGR888, orig_pixel_data_in[0]);
	    g_free(conv_pixel_data);
	  }
	  else {
	    // no palette conversion
	    memcpy(in_pixel_data[0],out_frames[0].pixel_data[0],in_frames[0].channel->height*in_frames[0].channel->rowstrides[0]);
	    g_free(out_frames[0].pixel_data[0]);
	  }
	  if (in_frames[0].channel->palette==LIVIDO_PALETTE_RGB32) {
	    // free RGB32 frame
	    g_free(in_frames[0].pixel_data[0]);
	    in_frames[0].pixel_data[0]=orig_pixel_data_in[0];
	    orig_pixel_data_in[0]=NULL;
	  }
	  if (in_frames[1].channel!=NULL&&in_frames[1].channel->palette==LIVIDO_PALETTE_RGB32) {
	    // free RGB32 frame
	    g_free(in_frames[1].pixel_data[0]);
	    in_frames[1].pixel_data[0]=orig_pixel_data_in[1];
	    orig_pixel_data_in[1]=NULL;
	  }
	}
	else {
	  if (in_frames[0].channel->palette==LIVIDO_PALETTE_RGB32) {
	    // convert in_frame BGR888-->RGB32
	    orig_pixel_data_in[0]=in_frames[0].pixel_data[0];
	    in_frames[0].pixel_data[0]=convert_frame_palette(in_frames[0].pixel_data[0], rowstride, height, LIVIDO_PALETTE_BGR888, LIVIDO_PALETTE_RGB32, NULL);
	  }
	  if (in_frames[1].channel!=NULL&&in_frames[1].channel->palette==LIVIDO_PALETTE_RGB32) {
	    // convert in_frame BGR888-->RGB32
	    orig_pixel_data_in[1]=in_frames[1].pixel_data[0];
	    in_frames[1].pixel_data[0]=convert_frame_palette(in_frames[1].pixel_data[0], rowstride, height, LIVIDO_PALETTE_BGR888, LIVIDO_PALETTE_RGB32, NULL);
	  }

	  (*fx->process)(instance,in_frames,NULL); // inplace

	  if (in_frames[0].channel->palette==LIVIDO_PALETTE_RGB32) {
	    conv_pixel_data=in_frames[0].pixel_data[0];
	    in_frames[0].pixel_data[0]=convert_frame_palette(in_frames[0].pixel_data[0], rowstride/3*4, height, LIVIDO_PALETTE_RGB32, LIVIDO_PALETTE_BGR888, orig_pixel_data_in[0]);
	    g_free(conv_pixel_data);
	    in_frames[0].pixel_data[0]=orig_pixel_data_in[0];
	    orig_pixel_data_in[0]=NULL;
	  }
	  if (in_frames[1].channel&&in_frames[1].channel->palette==LIVIDO_PALETTE_RGB32) {
	    // free RGB32 frame
	    g_free(in_frames[1].pixel_data[0]);
	    in_frames[1].pixel_data[0]=orig_pixel_data_in[1];
	    orig_pixel_data_in[1]=NULL;
	  }
	}
      }
    }
  }
  mainw->osc_block=FALSE;
}



void livido_deinit_all(void) {
  // deinit all except generators
  // this is called on ctrl-0 or on shutdown
  int i;
  gint idx;

  mainw->osc_block=TRUE;
  if (mainw->playing_file==-1&&rte_window!=NULL) {
    rtew_set_keygr(-1);
    mainw->rte_keys=-1;
    mainw->last_grabable_effect=-1;
  }

  for (i=0;i<FX_KEYS-1&&livido_key_valid (i)&&i<num_livido_effects;i++) {
    if (mainw->playing_file==-1&&rte_window!=NULL) rtew_set_keych(i,FALSE);
    if ((mainw->rte&(1<<i))) {
      if ((idx=key_to_instance[i][key_modes[i]])!=-1&&livido_instance[idx]!=NULL) {
	if (ienabled_in_channels(idx)>0) {
	  mainw->noswitch=TRUE;
	  livido_deinit_effect(i);
	  if (mainw->rte&(1<<i)) mainw->rte^=(1<<i);
	  if (rte_window!=NULL) {
	    rtew_set_keych(i,FALSE);
	    if (i==mainw->rte_keys) {
	      rtew_set_keygr(-1);
	      mainw->rte_keys=-1;
	    }
	  }
	  if (i==mainw->last_grabable_effect) mainw->last_grabable_effect=-1;
	}
      }
    }
  }
  mainw->osc_block=FALSE;
}

void livido_end(void) {
  int i;
  mainw->num_tr_applied=0; // don't bother with all that
  livido_deinit_all();
  for (i=0;i<num_livido_effects-num_modes;i++) {
    dlclose(livido_handles[i]);
    //g_free (livido_fx[i]);, etc
  }
}


gboolean livido_key_setmode (gint key, gint newmode) {
  gint oldmode;
  gint blend_file;
  gshort whentostop=mainw->whentostop;
  gboolean was_started=FALSE;
  gint idx,real_key;

  if (key==0) {
    if ((key=mainw->rte_keys)==-1) return FALSE;
  }
  else key--;

  real_key=key;

  oldmode=key_modes[key];

  if (newmode==-1) {
    if (oldmode==MAX_MODES_PER_KEY||key_to_fx[key][oldmode+1]==-1) {
      newmode=0;
    }
    else {
      newmode=key_modes[key]+1;
    }
  }

  if (newmode<0||newmode>=MAX_MODES_PER_KEY) return FALSE;

  if (key_to_fx[key][newmode]==-1) return FALSE;

  if (rte_window!=NULL) rtew_set_mode_radio(key,newmode);

  mainw->osc_block=TRUE;

  // TODO - block template channel changes

  if ((idx=key_to_instance[key][oldmode])!=-1&&livido_instance[idx]!=NULL){
    was_started=TRUE;
    if (ienabled_in_channels(idx)==2&&enabled_in_channels(key_to_fx[key][newmode])==2) {
      // transition --> transition, allow any bg generators to survive
      key=-key-1;
    }
  }

  if (oldmode!=newmode) {
    blend_file=mainw->blend_file;

    if (was_started&&ienabled_in_channels(idx)>0) {
      // not a generator
      livido_deinit_effect (key);
    }
    else if (enabled_in_channels(key_to_fx[key][newmode])==0) mainw->whentostop=NEVER_STOP; // when gen->gen, dont stop pb
    
    key_modes[real_key]=newmode;

    mainw->blend_file=blend_file;

    if (was_started) {
      if (!livido_init_effect(key,mainw->frame_image==NULL?0:gdk_pixbuf_get_rowstride(mainw->frame_image))) {
	// TODO - unblock template channel changes
	mainw->whentostop=whentostop;
	key=real_key;
	if (mainw->rte&(1<<key)) mainw->rte^=(1<<key);
	mainw->osc_block=FALSE;
	return FALSE;
       }
    }
    // TODO - unblock template channel changes
    mainw->whentostop=whentostop;
  }

  mainw->osc_block=FALSE;
  return TRUE;
}


int 
livido_get_id_for_hashname (const gchar *hashname) {
  int i;
  for (i=0;i<num_livido_effects;i++) {
    if (!strcmp(hashname,make_livido_hashname(i))) return i;
  }
  return -1;
}



gint 
livido_switch_keymode (gint key, gint mode, const gchar *hashname) {
  int oldkeymode=key_modes[--key];
  int id=livido_get_id_for_hashname (hashname);
  gboolean osc_block;
  gboolean has_gen=FALSE,has_non_gen=FALSE;

  // effect not found
  if (id==-1) return -1;

  if (key_to_fx[key][0]!=-1) {
    if (enabled_in_channels(key_to_fx[key][0])==0) has_gen=TRUE;
    else has_non_gen=TRUE;
  }

  if ((enabled_in_channels(id)==0&&has_non_gen)||(enabled_in_channels(id)>0&&has_gen)) return -2;

  osc_block=mainw->osc_block;
  mainw->osc_block=TRUE;
  if (key_to_instance[key][mode]!=-1) {
    key_modes[key]=mode;
    livido_deinit_effect(key);
    key_to_fx[key][mode]=id;
    livido_init_effect(key,mainw->frame_image==NULL?0:gdk_pixbuf_get_rowstride(mainw->frame_image));
    key_modes[key]=oldkeymode;
  }
  else key_to_fx[key][mode]=id;

  mainw->osc_block=osc_block;
  return 0;
}



void livido_swap_fg_bg (void) {
  gint key=fg_generator_key;
  gint mode=fg_generator_mode;

  if (key!=-1) {
    fg_generator_clip=-1;
  }
  fg_generator_key=bg_generator_key;
  fg_generator_mode=bg_generator_mode;
  if (fg_generator_key!=-1) {
    fg_generator_clip=mainw->current_file;
  }
  bg_generator_key=key;
  bg_generator_mode=mode;
}






int livido_key_getmode (gint key) {
  return key_modes[--key];
}



gchar *livido_keymode_get_name (gint key, gint mode) {
  // return value should be free'd after use
  key--;

  if (key<0||key>FX_KEYS-1||mode<0||mode>=MAX_MODES_PER_KEY) return g_strdup("");
  return livido_filter_get_name(key_to_fx[key][mode]);

}




gchar *livido_filter_get_name(gint idx) {
  // return value should be free'd after use
  gchar *name=g_strdup ("");
  livido_class_t *fx;

  if (idx==-1) return name;

  if ((fx=livido_fx[idx])==NULL) return name;
  if (fx->name==NULL) {
    g_free (name);
    return g_strdup (_("Unknown"));
  }
  g_free (name);
  return g_strdup_printf ("%s",fx->name);
}





gchar *livido_keymode_get_type (gint key, gint mode) {
  // return value should be free'd after use

  gchar *type=g_strdup ("");
  livido_class_t *fx;
  gint idx,inst_idx;

  key--;

  if (key<0||key>FX_KEYS-1||mode<0||mode>=MAX_MODES_PER_KEY) return type;

  if ((idx=key_to_fx[key][mode])==-1) return type;
  if ((fx=livido_fx[idx])==NULL) return type;

  mainw->osc_block=TRUE;

  if ((inst_idx=key_to_instance[key][mode])!=-1) {
    // return details for instance
    g_free(type);
    type=get_type (ienabled_in_channels(inst_idx),ienabled_out_channels(inst_idx));
    mainw->osc_block=FALSE;
  }
  else type=livido_filter_get_type(idx);
  return type;
}

gchar *livido_filter_get_type(gint idx) {
  // return value should be free'd after use
  return get_type(enabled_in_channels(idx),enabled_out_channels(idx));
}


gchar *get_type(gint in_channels, gint out_channels) {
  // return value should be free'd after use
  if (in_channels==0&&out_channels==1) return (g_strdup(_("generator")));
  if (in_channels==2&&out_channels==1) return (g_strdup(_("transition")));
  if (in_channels==1&&out_channels==1) return (g_strdup(_("effect")));
  if (in_channels==0&&out_channels==0) return (g_strdup(_("utility")));
  if (in_channels>2&&out_channels==1) return (g_strdup(_("compositor")));
  if (in_channels>0&&out_channels==0) return (g_strdup(_("tap")));
  if (in_channels>0&&out_channels>1) return (g_strdup(_("splitter")));
  return (g_strdup(_("unknown")));
}
  


GList *livido_get_all_names (gboolean add_type) {
  // remmeber to free list after use, if non-NULL
  GList *list=NULL;
  int i;

  for (i=0;i<num_livido_effects;i++) {
    if (add_type) list=g_list_append(list,(gpointer)(g_strdup_printf("%s (%s)",livido_fx[i]->name,livido_filter_get_type(i))));
    else list=g_list_append(list,(gpointer)g_strdup(livido_fx[i]->name));
  }
  return list;
}

gint livido_get_numeffects(void) {
  return num_livido_effects;
}


int livido_getmodespk (void) {
  return MAX_MODES_PER_KEY;
}


gint livido_bg_gen_key (void) {
  return bg_generator_key;
}

gint livido_fg_gen_key (void) {
  return fg_generator_key;
}

gint livido_bg_gen_mode (void) {
  return bg_generator_mode;
}

gint livido_fg_gen_mode (void) {
  return fg_generator_mode;
}

const livido_class_t *livido_keymode_get_class(gint key, gint mode) {
  // return value should NOT be g_free'd
  key--;
  if (!livido_keymode_valid(key+1,mode)) return NULL;
  return (const livido_class_t *)livido_fx[key_to_fx[key][mode]];
}

const livido_instance_t *livido_keymode_get_instance(gint key, gint mode) {
  // return value should NOT be g_free'd
  const livido_instance_t *inst;

  key--;
  if (!livido_keymode_valid(key+1,mode)) return NULL;
  mainw->osc_block=TRUE;
  if (key_to_instance[key][mode]==-1) {
    mainw->osc_block=FALSE;
    return NULL;
  }
  inst=livido_instance[key_to_instance[key][mode]];
  mainw->osc_block=FALSE;
  return inst;
}

gchar *livido_keymode_get_plugin_name(gint key, gint mode) {
  // return value should be g_free'd after use
  key--;
  if (!livido_keymode_valid(key+1,mode)) return g_strdup("");
  return g_strdup(plugin_names[key_to_fx[key][mode]]);
}


int livido_add_effectkey (gint key, gchar *hashname) {
  // we will add an effect at the next free slot for key, and return the slot number
  // if all slots are full, or the name was not found, we return -1
  // currently, generators and non-generators cannot be mixed on the same key (causes problems if the mode is switched)
  // in this case a -2 is returned
  gboolean has_gen=FALSE;
  gboolean has_non_gen=FALSE;

  int i,idx=livido_get_id_for_hashname(hashname);

  if (idx==-1) return -1;

  key--;

  for (i=0;i<MAX_MODES_PER_KEY;i++) {
    if (key_to_fx[key][i]!=-1) {
      if (enabled_in_channels(key_to_fx[key][i])==0) has_gen=TRUE;
      else has_non_gen=TRUE;
    }
    else {
      if ((enabled_in_channels(idx)==0&&has_non_gen)||(enabled_in_channels(idx)>0&&has_gen)) return -2;
      key_to_fx[key][i]=idx;
      return i;
    }
  }
  return -1;
}


gboolean livido_delete_effectkey (gint key, gint mode) {
  // delete the effect binding for key/mode and move higher numbered slots down
  // also moves the active mode if applicable
  // returns FALSE if there was no effect bound to key/mode

  int oldkeymode=key_modes[--key];
  gint orig_mode=mode;
  gboolean was_started=FALSE;

  if (key_to_fx[key][mode]==-1) return FALSE;

  for (;mode<MAX_MODES_PER_KEY;mode++) {

    mainw->osc_block=TRUE;
    if (mode==MAX_MODES_PER_KEY-1||key_to_fx[key][mode+1]==-1) {
      if (key_to_instance[key][mode]!=-1) {
	was_started=TRUE;
	key_modes[key]=mode;
	livido_deinit_effect(key);
	key_modes[key]=oldkeymode;
      }

      key_to_fx[key][mode]=-1;

      if (mode==orig_mode&&key_modes[key]==mode) {
	key_modes[key]=0;
	if (was_started) {
	  if (key_to_fx[key][0]!=-1) livido_init_effect(key,mainw->frame_image==NULL?0:gdk_pixbuf_get_rowstride(mainw->frame_image));
	  else if (mainw->rte&(1<<key)) mainw->rte^=(1<<key);
	}
      }
      break; // quit the loop
    }
    else {
      livido_switch_keymode(key+1,mode,make_livido_hashname(key_to_fx[key+1][mode+1]));
    }
  }

  mainw->osc_block=FALSE;
  if (key_modes[key]>orig_mode) key_modes[key]--;

  return TRUE;
}




gchar *make_livido_hashname(int fx_idx) {
  const livido_class_t *class;
  const gchar *plugin_name;

  if (fx_idx<0||fx_idx>=num_livido_effects) return g_strdup("");

  plugin_name=plugin_names[fx_idx];
  class=livido_fx[fx_idx];

  return g_strconcat(plugin_name,class->name,class->author,NULL);
}
